package at.krixec.rosary

import android.app.LocaleManager
import android.graphics.PathMeasure
import android.os.Build
import android.os.Bundle
import android.os.LocaleList
import androidx.activity.compose.setContent
import androidx.activity.enableEdgeToEdge
import androidx.appcompat.app.AppCompatActivity
import androidx.appcompat.app.AppCompatDelegate
import androidx.compose.animation.core.Animatable
import androidx.compose.animation.core.EaseInOutCubic
import androidx.compose.animation.core.LinearEasing
import androidx.compose.animation.core.repeatable
import androidx.compose.animation.core.tween
import androidx.compose.foundation.Canvas
import androidx.compose.foundation.background
import androidx.compose.foundation.gestures.detectDragGestures
import androidx.compose.foundation.gestures.detectTapGestures
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.PaddingValues
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.offset
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.material3.Button
import androidx.compose.material3.ButtonDefaults
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.derivedStateOf
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableFloatStateOf
import androidx.compose.runtime.mutableIntStateOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.geometry.Offset
import androidx.compose.ui.geometry.Rect
import androidx.compose.ui.geometry.Size
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.Path
import androidx.compose.ui.graphics.Shadow
import androidx.compose.ui.graphics.asAndroidPath
import androidx.compose.ui.graphics.drawscope.Stroke
import androidx.compose.ui.input.pointer.pointerInput
import androidx.compose.ui.platform.LocalConfiguration
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.text.AnnotatedString
import androidx.compose.ui.text.SpanStyle
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.buildAnnotatedString
import androidx.compose.ui.text.drawText
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.rememberTextMeasurer
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.text.withStyle
import androidx.compose.ui.unit.Dp
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.core.os.LocaleListCompat
import androidx.core.view.WindowCompat
import androidx.core.view.WindowInsetsCompat
import androidx.core.view.WindowInsetsControllerCompat
import at.krixec.rosary.ui.theme.Bead
import at.krixec.rosary.ui.theme.Cross
import at.krixec.rosary.ui.theme.Orange
import at.krixec.rosary.ui.theme.RosariumTheme
import at.krixec.rosary.ui.theme.SelectedBead
import at.krixec.rosary.ui.theme.Shade
import kotlinx.coroutines.flow.first
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import java.time.LocalDate
import androidx.compose.ui.graphics.PathMeasure as PM


// Google is a joke
// https://stackoverflow.com/a/74413010
class MainActivity : AppCompatActivity() {

    private var isFirstTime = false

    override fun onCreate( savedInstanceState: Bundle? ) {
        super.onCreate( savedInstanceState )

        val insetsController =
            WindowCompat.getInsetsController(
                window,
                window.decorView
            )

        insetsController.apply {
            hide( WindowInsetsCompat.Type.statusBars() )
            hide( WindowInsetsCompat.Type.navigationBars() )
            systemBarsBehavior = WindowInsetsControllerCompat.BEHAVIOR_SHOW_TRANSIENT_BARS_BY_SWIPE
        }
        WindowCompat.setDecorFitsSystemWindows( window, false )

        enableEdgeToEdge()
        setContent {

            val context = LocalContext.current
            val store = DataSave( context )

            // asynchronous preferences are retarded
            val immaculate = runBlocking {
                store.isImmaculate.first()
            }
            if ( immaculate )
                isFirstTime = true

            // I set the language to latin if it isn't either english or german
            val currentLang = LocalConfiguration.current.locales[ 0 ].toString().substring( 0, 2 )

            if ( currentLang !in listOf( "de", "en", "lt" ) ) {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU) {
                    this.getSystemService(LocaleManager::class.java)
                        .applicationLocales = LocaleList.forLanguageTags("lt")
                } else {
                    AppCompatDelegate.setApplicationLocales(
                        LocaleListCompat.forLanguageTags("lt")
                    )
                }
            } else if ( isFirstTime ) {
                if ( Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU ) {
                    this.getSystemService( LocaleManager::class.java )
                        .applicationLocales = LocaleList.forLanguageTags( currentLang )
                } else {
                    AppCompatDelegate.setApplicationLocales(
                        LocaleListCompat.forLanguageTags( currentLang )
                    )
                }
            }

            RosariumTheme {
                App( isFirstTime, store )
            }
        }
    }
}

@Composable
private fun App( immaculate: Boolean, store: DataSave ) {

    val scope = rememberCoroutineScope()

    val showInfoPage  = remember { mutableStateOf( immaculate ) }
    val configuration = LocalConfiguration.current
    val screenWidth   = configuration.screenWidthDp.dp
    val screenHeight  = configuration.screenHeightDp.dp
    val density  = LocalDensity.current.density
    val progress = remember { Animatable(
        if ( immaculate ) 1.8f else 0f
    ) }
    val textMeasurer = rememberTextMeasurer()

    @Composable
    fun InfoPage( showPage: Boolean, onChange: ( Boolean ) -> Unit ) {

        if ( showPage ) {
            Box(
                modifier = Modifier
                    .fillMaxSize()
                    .background( Shade )
                    .pointerInput( Unit ) {

                        detectTapGestures {
                            scope.launch {

                                store.setNotImmaculate()

                                onChange( false )
                                progress.animateTo(
                                    0f,
                                    animationSpec = repeatable(
                                        iterations = 1,
                                        animation = tween(
                                            durationMillis = 1200,
                                            easing = EaseInOutCubic
                                        )
                                    )
                                )
                            }
                        }
                    }
            ) {

                Column(
                    modifier = Modifier
                        .padding( top = 25.dp )
                        .fillMaxSize(),
                    horizontalAlignment = Alignment.CenterHorizontally
                ) {

                    Text(
                        color = Orange,
                        fontSize = 22.sp,
                        fontWeight = FontWeight.ExtraBold,
                        textAlign = TextAlign.Center,
                        text = stringResource( R.string.change_lang ) + " →"
                    )

                }
                Column(
                    modifier = Modifier
                        .padding( bottom = 160.dp )
                        .fillMaxSize(),
                    verticalArrangement = Arrangement
                        .spacedBy( 120.dp, Alignment.Bottom ),
                    horizontalAlignment = Alignment.CenterHorizontally
                ) {

                    Text(
                        color = Orange,
                        fontSize = 22.sp,
                        fontWeight = FontWeight.ExtraBold,
                        textAlign = TextAlign.Center,
                        text = stringResource( R.string.tap_to_proceed ) + "\n►►►"
                    )

                    Text(
                        color = Orange,
                        fontSize = 22.sp,
                        fontWeight = FontWeight.ExtraBold,
                        textAlign = TextAlign.Center,
                        text = stringResource( R.string.swipe_to_go_back ) + "\n◄◄◄"
                    )
                }
            }
        }
    }

    Rosary( screenWidth, screenHeight )
    InfoPage( showInfoPage.value, onChange = { showInfoPage.value = it } )
    SettingsMenu( screenWidth, density )


    Button(
        modifier = Modifier
            .offset( -( 16.dp ), -( 10f.dp ) )
            .size( 86.dp ),
        contentPadding = PaddingValues( 1.dp ),
        colors = ButtonDefaults.buttonColors( Color.Transparent ),

        onClick = {
            scope.launch {
                progress.animateTo(
                    1.8f,
                    animationSpec = repeatable(
                        iterations = 1,
                        animation = tween(
                            durationMillis = 1200,
                            easing = EaseInOutCubic
                        )
                    )
                )
                showInfoPage.value = true
            }
        }
    ) {}

    Canvas(
        modifier = Modifier
            .background( Color.Transparent )
            .size(screenWidth, screenHeight)
    ) {
        drawLine(
            color = Color.Gray,
            start = Offset( 85f, 0f ),
            end   = Offset( 85f, 50f + 36.dp.toPx() - ( 50f + 40.dp.toPx() ) * progress.value ),
            strokeWidth = 16f
        )
        drawCircle(
            radius = 16.dp.toPx(),
            color = Bead,
            center = Offset( 85f, 50f + 16.dp.toPx() - ( 50f + 16.dp.toPx() ) * progress.value )
        )
        drawCircle(
            radius = 4.dp.toPx(),
            color = Color.Gray,
            center = Offset( 85f, 50f + 36.dp.toPx() - ( 50f + 16.dp.toPx() ) * progress.value )
        )
        drawText(
            textMeasurer = textMeasurer,
            text = "?",
            topLeft = Offset( 85f - 26.sp.toPx() / 4, 54f - ( 54f + 16.dp.toPx() ) * progress.value ),
            style = TextStyle.Default.copy(
                fontSize = 26.sp,
                fontWeight = FontWeight.Bold,
                color = Color.LightGray,
                shadow = Shadow(
                    color = Color.Black,
                    offset = Offset( 3f, 3f )
                )
            )
        )
    }
}

@Composable
private fun SettingsMenu( screenWidth: Dp, density: Float ) {

    val circle = Path()
    val scope = rememberCoroutineScope()
    val context = LocalContext.current
    val textMeasurer = rememberTextMeasurer()
    var currentLang = LocalConfiguration.current.locales[ 0 ].toString()

    val screenWidthInPx = screenWidth.value * density

    val progress  = remember { Animatable( 0f ) }
    var animating = remember { false }
    //val radius = 400f
    val radius = screenWidthInPx / 2.8f
    var circleCenter = remember { Pair( 0f, 0f ) }

    circle.reset()
    circle.addOval(
        Rect(
            Offset( screenWidthInPx - radius / 2, 0f - radius / 2 ),
            Size( radius, radius / 1.2f )
        )
    )

    fun getFlag( currentLang: String ): String {

        return if ( !animating ) {
            when( currentLang ) {
                "lt" -> "\uD83C\uDDFB\uD83C\uDDE6"
                "de" -> "\uD83C\uDDE6\uD83C\uDDF9"
                "en" -> "\uD83C\uDDEC\uD83C\uDDE7"
                else -> "\uD83C\uDDFB\uD83C\uDDE6"
            }
        } else
            ""
    }

    fun getNextLang( currentLang: String ): String {

        return when( currentLang ) {
            "lt" -> "de"
            "de" -> "en"
            "en" -> "lt"
            else -> "lt"
        }
    }

    val flagPos = FloatArray( 2 )
    val flagTan = FloatArray( 2 )
    PathMeasure().apply {
        setPath(circle.asAndroidPath(), false)
        getPosTan(
            length * ( 2 * ( 1f / 3f ) - 0.03f ),
            flagPos,
            flagTan
        )
    }
    circleCenter = Pair( flagPos[ 0 ], flagPos[ 1 ] )

    Canvas(
        modifier = Modifier
            .background( Color.Transparent )
            .fillMaxSize()
    ) {
        drawPath(
            circle,
            style = Stroke( 16f ),
            color = Color.Gray
        )

        for ( bead in 0..2 ) {

            for ( knot in -1..1 ) {
                val pos = FloatArray( 2 )
                val tan = FloatArray( 2 )

                PathMeasure().apply {
                    setPath(circle.asAndroidPath(), false)
                    getPosTan(
                        length * ( progress.value + ( bead + ( knot * 0.17f ) ) * ( 1f / 3f ) - 0.03f ),
                        pos,
                        tan
                    )
                }

                if ( knot == 0 ) {
                    drawCircle(
                        radius = 16.dp.toPx(),
                        color = Bead,
                        center = Offset( pos[ 0 ], pos[ 1 ] )
                    )

                } else {
                    drawCircle(
                        radius = 8.dp.toPx(),
                        color = Color.Gray,
                        center = Offset( pos[0], pos[1] )
                    )
                }
            }
        }

        val flag = getFlag( currentLang )
        val style = TextStyle.Default.copy(
            fontSize = 20.sp,
            shadow = Shadow( color = Color.Black, offset = Offset( 5f, 8f ) )
        )
        val textLayoutResult = textMeasurer.measure( flag, style )

        drawText(
            textMeasurer = textMeasurer,
            text = flag,
            style = style,
            topLeft = Offset(
                circleCenter.first - textLayoutResult.size.width / 2 - 4,
                circleCenter.second - textLayoutResult.size.height / 2 - 4
            )
        )
    }

    val pos = FloatArray( 2 )
    val tan = FloatArray( 2 )

    PathMeasure().apply {
        setPath( circle.asAndroidPath(), false )
        getPosTan( length * 0.55f,
            pos,
            tan
        )
    }

    Button(
        modifier = Modifier
            //.offset( ( pos[ 0 ] / density ).dp, ( pos[ 1 ] / density ).dp )
            .offset( ( circleCenter.first / density ).dp - 43.dp, ( circleCenter.second / density ).dp - 43.dp )
            .size( 86.dp ),
        contentPadding = PaddingValues( 1.dp ),

        onClick = {
            scope.launch {
                if ( !animating ) {
                    animating = true
                    progress.animateTo(
                        progress.value + ( 1f / 3f ),
                        animationSpec = repeatable(
                            iterations = 1,
                            animation = tween(
                                durationMillis = 1000,
                                easing = LinearEasing
                            )
                        )
                    )

                    progress.snapTo( 0f )

                    val lang = getNextLang( currentLang )
                    currentLang = lang

                    // locales in Android give me turbo cancer
                    if ( Build.VERSION.SDK_INT >= Build.VERSION_CODES.TIRAMISU ) {
                        context.getSystemService( LocaleManager::class.java )
                            .applicationLocales = LocaleList.forLanguageTags( lang )
                    } else {
                        AppCompatDelegate.setApplicationLocales(
                            LocaleListCompat.forLanguageTags(
                                lang
                            )
                        )
                    }

                    animating = false
                }
            }
        },
        colors = ButtonDefaults.buttonColors(
            containerColor = Color.Transparent
        )
    ) { }
}

/* https://www.baeldung.com/java-determine-easter-date-specific-year */
fun getEaster( year: Int ): LocalDate? {
    val a = year % 19
    val b = year / 100
    val c = year % 100
    val d = b / 4
    val e = b % 4
    val f = ( b + 8 ) / 25
    val g = ( b - f + 1 ) / 3
    val h = ( 19 * a + b - d - g + 15 ) % 30
    val i = c / 4
    val k = c % 4
    val l = ( 2 * e + 2 * i - h - k + 32 ) % 7
    val m = ( a + 11 * h + 22 * l ) / 451
    val t = h + l - 7 * m + 114
    val n = t / 31
    val o = t % 31

    return LocalDate.of( year, n, o + 1 )
}

@Composable
fun getHolyMystery( decade: Int, bead: Int, german: Boolean ): String {

    if ( decade == 0 && german )
        return when( bead ) {
            3 -> stringResource( R.string.ave_maria_ger_1 )
            4 -> stringResource( R.string.ave_maria_ger_2 )
            5 -> stringResource( R.string.ave_maria_ger_3 )
            else -> "\n"
        }

    val date = LocalDate.now()
    val dayOfWeek = date.dayOfWeek.value

    val easter = getEaster( date.year )
    val lent = easter?.minusDays( 40 )
    val christmas = LocalDate.of( date.year, 12, 25 )
    val startOfAdvent = christmas.minusDays( ( 21 + christmas.dayOfWeek.value ).toLong() )

    // Monday -> 1, Sunday -> 7
    if ( dayOfWeek == 1 || dayOfWeek == 6 || date == christmas ||
        ( dayOfWeek == 7 && ( date >= startOfAdvent && date < christmas ) ) )
        return when( decade ) {
            1 -> stringResource( R.string.gaudiosa_1 )
            2 -> stringResource( R.string.gaudiosa_2 )
            3 -> stringResource( R.string.gaudiosa_3 )
            4 -> stringResource( R.string.gaudiosa_4 )
            5 -> stringResource( R.string.gaudiosa_5 )
            else -> ""
        }
    else if ( dayOfWeek == 4 )
        return when( decade ) {
            1 -> stringResource( R.string.luminosa_1 )
            2 -> stringResource( R.string.luminosa_2 )
            3 -> stringResource( R.string.luminosa_3 )
            4 -> stringResource( R.string.luminosa_4 )
            5 -> stringResource( R.string.luminosa_5 )
            else -> ""
        }
    else if ( dayOfWeek == 2 || dayOfWeek == 5 ||
        ( dayOfWeek == 7 && ( date >= lent && date < easter ) ) )
        return when( decade ) {
            1 -> stringResource( R.string.dolorosa_1 )
            2 -> stringResource( R.string.dolorosa_2 )
            3 -> stringResource( R.string.dolorosa_3 )
            4 -> stringResource( R.string.dolorosa_4 )
            5 -> stringResource( R.string.dolorosa_5 )
            else -> ""
        }
    else if ( dayOfWeek == 3 || dayOfWeek == 7 )
        return when( decade ) {
            1 -> stringResource( R.string.gloriosa_1 )
            2 -> stringResource( R.string.gloriosa_2 )
            3 -> stringResource( R.string.gloriosa_3 )
            4 -> stringResource( R.string.gloriosa_4 )
            5 -> stringResource( R.string.gloriosa_5 )
            else -> ""
        }

    return ""
}

@Composable
private fun Rosary( screenWidth: Dp, screenHeight: Dp ) {

    val path = Path()

    var prayer by remember { mutableIntStateOf( 0 ) }
    var decade by remember { mutableIntStateOf( 0 ) }
    var bead   by remember { mutableIntStateOf( 0 ) }

    // I've bound the actual decade to an animation, little workaround
    var bonusDecade by remember { mutableIntStateOf( 0 ) }


    @Composable
    fun RosaryLoop() {

        var totalDrag by remember { mutableFloatStateOf( 0f ) }
        var hasMoved  = false
        var animating = false

        // who could take compose seriously with stuff like this???
        val scope = rememberCoroutineScope()

        // left -> 0, middle -> 1, right -> 2
        var current by remember { mutableIntStateOf( 1 ) }

        fun getTargetValue( right: Boolean ): Float {
            return if ( current == 1 )
                if ( right )
                    -0.8f
                else
                    1.35f
            else
                0.2f
        }

        @Composable
        fun getPrayer(): AnnotatedString {

            val currentLang = LocalConfiguration.current.locales[ 0 ].toString()

            @Composable
            fun buildHailMary(): AnnotatedString {

                val realDecade = decade + bonusDecade

                val mystery = getHolyMystery( realDecade, bead, currentLang == "de" )
                return if ( currentLang != "de" )
                    buildAnnotatedString {
                        withStyle( style = SpanStyle( color = Orange, fontWeight = FontWeight.Bold ) ) {
                            append( mystery )
                        }
                        append( stringResource( R.string.ave_maria_1 ) )

                        append( stringResource( R.string.ave_maria_2 ) )
                    }
                else
                    buildAnnotatedString {
                        append( stringResource( R.string.ave_maria_1 ) )

                        withStyle( style = SpanStyle( color = Orange, fontWeight =  FontWeight.Bold ) ) {
                            append( mystery )
                        }
                        append( stringResource( R.string.ave_maria_2 ) )
                    }
            }

            return when ( prayer ) {
                0 -> buildAnnotatedString { append( stringResource( R.string.credo ) ) }
                1 -> buildAnnotatedString { append( stringResource( R.string.pater_noster ) ) }
                2 -> buildAnnotatedString { append( stringResource( R.string.gloria_patri ) ) }
                3 -> buildAnnotatedString { append( stringResource( R.string.fatima ) ) }
                4 -> buildAnnotatedString { append( stringResource( R.string.credo_2 ) ) }
                5 -> buildAnnotatedString { append( stringResource( R.string.salve_regina ) ) }

                else -> {
                    buildHailMary()
                }
            }
        }

        var fullyShown   = remember { false }
        val progress     = remember { Animatable( 0.2f ) }
        val progress2    = remember { Animatable( 0.2f ) }
        val pathMeasure  = remember { PM() }
        val animatedPath = remember {
            derivedStateOf {
                val destination = Path()
                pathMeasure.setPath( path, false )

                // this could probably be simpler
                pathMeasure.getSegment(
                    0f,
                    pathMeasure.length + progress2.value * pathMeasure.length - pathMeasure.length * 0.2f,
                    destination
                )
                destination
            }
        }

        fun goRight() {

            if ( !animating ) {

                if ( decade == 0 ) {

                    if ( bead < 7 ) {
                        bead += 1
                        prayer = when ( bead ) {
                            1 -> 4
                            2 -> 1
                            6 -> 2
                            7 -> 1

                            else -> -1
                        }
                    } else {

                        scope.launch {
                            animating = true
                            prayer = -1
                            bonusDecade = 1
                            progress2.animateTo(
                                -1f,
                                animationSpec = repeatable(
                                    iterations = 1,
                                    animation = tween(
                                        durationMillis = 1500,
                                        easing = LinearEasing
                                    )
                                )
                            )
                            bonusDecade = 0
                            decade += 1
                            bead   = 0
                            progress2.snapTo( -0.5f )
                            progress2.animateTo(
                                -1f,
                                animationSpec = repeatable(
                                    iterations = 1,
                                    animation = tween(
                                        durationMillis = 1000,
                                        easing = LinearEasing
                                    )
                                )
                            )
                            fullyShown = true
                            animating = false
                        }
                    }
                } else {
                    if ( decade == 5 && bead == 10 ) {

                        prayer = 5
                        bead += 1

                    } else if ( decade == 5 && bead == 11 ) {

                        scope.launch {

                            animating = true
                            fullyShown = false
                            prayer = 0
                            progress2.animateTo(
                                0.2f,
                                animationSpec = repeatable(
                                    iterations = 1,
                                    animation = tween(
                                        durationMillis = 1500,
                                        easing = LinearEasing
                                    )
                                )
                            )
                            bead   = 0
                            decade = 0
                            progress2.snapTo( -1f )
                            progress2.animateTo(
                                0.2f,
                                animationSpec = repeatable(
                                    iterations = 1,
                                    animation = tween(
                                        durationMillis = 1000,
                                        easing = LinearEasing
                                    )
                                )
                            )

                            animating = false
                        }
                    } else if ( ( current == 1 && bead >= 10 )
                        || ( current != 1 && bead >= 4 )
                    ) {

                        scope.launch {

                            animating = true
                            if ( current != 1 ) {
                                progress.snapTo( 1.34f )
                                bead = 0
                                prayer = -1
                            } else {
                                bead = 3
                                prayer = 2
                            }

                            if ( current == 1 )
                                if ( decade < 5 )
                                    decade += 1
                                else
                                    decade = 0

                            if ( decade != 0 ) {
                                progress.animateTo(
                                    getTargetValue( true ),
                                    animationSpec = repeatable(
                                        iterations = 1,
                                        animation = tween(
                                            durationMillis = 2000,
                                            easing = LinearEasing
                                        )
                                    )
                                )

                                if ( current < 2 )
                                    current += 1
                                else {
                                    current = 1
                                }
                            } else {
                                fullyShown = false
                                bead = 0
                                progress2.animateTo(
                                    0.2f,
                                    animationSpec = repeatable(
                                        iterations = 1,
                                        animation = tween(
                                            durationMillis = 1500,
                                            easing = LinearEasing
                                        )
                                    )
                                )
                            }
                            animating = false
                        }
                    } else {
                        bead += 1
                        if ( bead == 10 )
                            prayer = 3
                        if ( current != 1 )
                            prayer = 1
                    }
                }
            }
        }

        fun goLeft() {
            if ( decade == 0 && bead == 0 ) {
                scope.launch {

                    animating = true
                    prayer = 5
                    bonusDecade = 5
                    progress2.animateTo(
                        -1f,
                        animationSpec = repeatable(
                            iterations = 1,
                            animation = tween(
                                durationMillis = 1500,
                                easing = LinearEasing
                            )
                        )
                    )
                    bonusDecade = 0
                    decade = 5
                    bead   = 11
                    progress2.snapTo( -0.5f )
                    progress2.animateTo(
                        -1f,
                        animationSpec = repeatable(
                            iterations = 1,
                            animation = tween(
                                durationMillis = 1000,
                                easing = LinearEasing
                            )
                        )
                    )
                    animating  = false
                    fullyShown = true
                }
            } else if ( decade == 1 && bead == 0 ) {

                scope.launch {

                    animating  = true
                    fullyShown = false
                    prayer = 1
                    progress2.animateTo(
                        0.2f,
                        animationSpec = repeatable(
                            iterations = 1,
                            animation = tween(
                                durationMillis = 1500,
                                easing = LinearEasing
                            )
                        )
                    )
                    decade = 0
                    bead   = 7
                    progress2.snapTo( -1f )
                    progress2.animateTo(
                        0.2f,
                        animationSpec = repeatable(
                            iterations = 1,
                            animation = tween(
                                durationMillis = 1000,
                                easing = LinearEasing
                            )
                        )
                    )

                    animating = false
                }

            } else if ( ( current == 1 && bead <= 0 )
                || ( current != 1 && bead <= 3 )
            ) {
                scope.launch {
                    animating = true
                    if ( current != 1 ) {
                        progress.snapTo( -0.78f )
                        bead = 10
                        prayer = 3
                    } else {
                        bead = 4
                        prayer = 1
                    }

                    if ( current == 1 )
                        if ( decade > 0 )
                            decade -= 1
                        else {
                            decade = 5
                        }

                    progress.animateTo(
                        getTargetValue( false ),
                        animationSpec = repeatable(
                            iterations = 1,
                            animation = tween(
                                durationMillis = 2000,
                                easing = LinearEasing
                            )
                        )
                    )

                    if ( current > 0 )
                        current -= 1
                    else {
                        current = 1
                    }
                    animating = false
                }
            } else {
                bead -= 1
                if ( decade != 0 )
                    prayer = -1
                else {
                    if ( bead < 7 ) {
                        prayer = when ( bead ) {
                            0 -> 0
                            1 -> 4
                            2 -> 1
                            6 -> 2
                            //7 -> 1

                            else -> -1
                        }
                    }
                }
                if ( decade == 5 && bead == 10 )
                    prayer = 3
                else if ( current != 1 && decade != 0 )
                    prayer = 2
            }
        }

        // invisible box to catch touch input
        Box(
            modifier = Modifier
                .fillMaxSize()
                .pointerInput( Unit ) {

                    detectTapGestures {
                        goRight()
                    }
                }
                .pointerInput( Unit ) {
                    detectDragGestures(
                        onDragStart = {
                            totalDrag = 0f
                            hasMoved = false
                        }
                    ) { change, dragAmount ->
                        change.consume()

                        totalDrag += dragAmount.x

                        if ( !hasMoved && !animating && totalDrag > 20) {
                            hasMoved = true
                            goRight()
                        } else if ( !hasMoved && !animating && totalDrag < -20) {
                            hasMoved = true
                            goLeft()
                        }
                    }
                }
        )

        val density = LocalDensity.current.density

        val screenWidthInPx  = screenWidth.value * density
        val screenHeightInPx = screenHeight.value * density

        Text(
            getPrayer(),
            modifier = Modifier
                .fillMaxSize()
                .padding( 20.dp, 100.dp ),
            textAlign = TextAlign.Center,
            color = Color.LightGray,
            fontSize = 18.sp,
            fontWeight = FontWeight.ExtraBold
        )

        Canvas(
            modifier = Modifier
                .background( Color.Transparent )
                .size( screenWidth, screenHeight / 1.5f )
        ) {

            if ( decade != 0 ) {
                val x0 = -100f
                val y0 = screenHeightInPx / 1.2f
                val y1 = screenHeightInPx / 1.6f
                val y2 = screenHeightInPx / 0.9f
                val y3 = screenHeightInPx / 1.2f

                path.reset()
                path.moveTo( x0, if ( fullyShown ) y0 else y0 + screenHeightInPx * ( progress2.value + 1f ) )
                path.cubicTo(
                    x1 = screenWidthInPx / 2f,
                    y1 = if ( fullyShown ) y1 else y1 + screenHeightInPx / 3 * ( progress2.value + 1f ) ,
                    x2 = screenWidthInPx / 2f,
                    y2 = if ( fullyShown ) y2 else y2 + screenHeightInPx / 3 * ( progress2.value + 1f ) ,
                    x3 = screenWidthInPx + 100f,
                    y3 = if ( fullyShown ) y3 else y3 + screenHeightInPx / 3 * ( progress2.value + 1f ) ,
                )

                drawPath(
                    color = Color.Gray,
                    path  = path,
                    style = Stroke(
                        width = 8.dp.toPx()
                    )
                )

                // side pater nosters
                for ( i in 0..5 ) {

                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )
                    val mod = if ( i <= 2 ) -1 else 1

                    PathMeasure().apply {
                        setPath( path.asAndroidPath(), false )
                        getPosTan( length * ( progress.value + mod + ( i + 2 ) * 0.045f ),
                            pos,
                            tan
                        )
                    }

                    val color = if ( prayer == 1 ) SelectedBead else Bead
                    if ( i % 2 != 0 && i < 3 ) {
                        drawCircle(
                            radius = 16.dp.toPx(),
                            color  = color,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    } else if ( i % 2 == 0 && i < 3 ) {
                        drawCircle(
                            radius = 8.dp.toPx(),
                            color  = Color.Gray,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    } else if ( i % 2 != 0 ) {
                        drawCircle(
                            radius = 8.dp.toPx(),
                            color  = Color.Gray,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    } else {
                        drawCircle(
                            radius = 16.dp.toPx(),
                            color  = color,
                            center = Offset( pos[ 0 ], pos[ 1 ] )
                        )
                    }
                }

                // grey knots
                for ( i in 0..1 ) {

                    val mod = if ( i == 0 ) -1f else 14.15f
                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )

                    PathMeasure().apply {
                        setPath( path.asAndroidPath(), false )
                        getPosTan( length * ( progress.value + mod * 0.048f ),
                            pos,
                            tan
                        )
                    }

                    drawCircle(
                        radius = 8.dp.toPx(),
                        color  = Color.Gray,
                        center = Offset( pos[ 0 ], pos[ 1 ] )
                    )

                }
                for ( i in 0..9 ) {

                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )

                    PathMeasure().apply {
                        setPath( path.asAndroidPath(), false )
                        getPosTan( length * ( progress.value + i * 0.07f ),
                            pos,
                            tan
                        )
                    }
                    val color = if ( bead == i && prayer == -1 ) SelectedBead else Bead
                    drawCircle(
                        radius = 16.dp.toPx(),
                        color = color,
                        center = Offset( pos[ 0 ], pos[ 1 ] )
                    )
                }
            } else {
                val x0 = screenWidthInPx + 220f
                val y0 = screenHeightInPx / 1.8f

                path.reset()
                path.moveTo( x0, y0 )
                path.cubicTo(
                    x1 = screenWidthInPx  / 2f,
                    y1 = screenHeightInPx / 1.6f,
                    x2 = screenWidthInPx  / 1.4f,
                    y2 = screenHeightInPx / 1.4f,
                    x3 = screenWidthInPx  / 1.6f,
                    y3 = screenHeightInPx / 1.3f
                )

                drawPath(
                    color = Color.Gray,
                    path = animatedPath.value,
                    style = Stroke(
                        width = 8.dp.toPx()
                    )
                )

                for ( i in 0..5 ) {

                    val pos = FloatArray( 2 )
                    val tan = FloatArray( 2 )

                    when ( i ) {
                        0 -> {

                            PathMeasure().apply {
                                setPath( path.asAndroidPath(), false )
                                getPosTan(
                                    length * ( progress2.value + 0.8f ),
                                    pos,
                                    tan
                                )
                            }

                            val color = Cross
                            val girth = 64f
                            drawLine(
                                color = color,
                                start = Offset( pos[ 0 ] - 120f - girth / 2, pos[ 1 ] ),
                                end   = Offset( pos[ 0 ] + 100f - girth / 2, pos[ 1 ] + 130f ),
                                strokeWidth = girth
                            )
                            drawLine(
                                color = color,
                                start = Offset(pos[ 0 ] + 0f, pos[ 1 ] - 10f ),
                                end   = Offset(pos[ 0 ] - 180f, pos[ 1 ] + 300f),
                                strokeWidth = girth
                            )
                        }
                        1 -> {
                            // if the following numbers seem arbitrary, that's because they are
                            for ( j in -1..1 ) {

                                PathMeasure().apply {
                                    setPath( path.asAndroidPath(), false )
                                    getPosTan( length * ( progress2.value + j * 0.076f + 0.13f ),
                                        pos,
                                        tan
                                    )
                                }

                                val color = if ( bead == 7 ) SelectedBead else Bead
                                if ( j != 0 )
                                    drawCircle(
                                        radius = 8.dp.toPx(),
                                        color  = Color.Gray,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                                else
                                    drawCircle(
                                        radius = 16.dp.toPx(),
                                        color  = color,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                            }
                            PathMeasure().apply {
                                setPath( path.asAndroidPath(), false )
                                getPosTan( length * ( progress2.value + 0.01f ),
                                    pos,
                                    tan
                                )
                            }

                        }
                        in 2..4 -> {
                            PathMeasure().apply {
                                setPath( path.asAndroidPath(), false )
                                getPosTan( length * ( progress2.value + i * 0.105f + 0.08f ),
                                    pos,
                                    tan
                                )
                            }

                            val color = if ( bead == 7 - i ) SelectedBead else Bead
                            drawCircle(
                                radius = 16.dp.toPx(),
                                color  = color,
                                center = Offset(pos[ 0 ], pos[ 1 ])
                            )
                        }
                        else -> {
                            for ( j in -1..1 ) {

                                PathMeasure().apply {
                                    setPath( path.asAndroidPath(), false )
                                    getPosTan(
                                        length * ( progress2.value + i * 0.125f + j * 0.077f + 0.035f ),
                                        pos,
                                        tan
                                    )
                                }

                                val color = if ( bead == 2 ) SelectedBead else Bead
                                if ( j != 0 )
                                    drawCircle(
                                        radius = 8.dp.toPx(),
                                        color  = Color.Gray,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                                else
                                    drawCircle(
                                        radius = 16.dp.toPx(),
                                        color  = color,
                                        center = Offset( pos[ 0 ], pos[ 1 ] )
                                    )
                            }
                        }
                    }
                }
            }
        }
    }

    RosaryLoop()
}
